# ==================== #
# CALIBRATION MEASURES #
# ==================== #

# Loss Functions for Binary Class Probability Estimation
# ------------------------------------------------------

# Test data: y_1, y_2, ..., y_n \in {0,1}
# Estimated Probabilities: p_1, p_2, ..., p_n \in [0,1]

# INPUTS
# ------
# y: real class (array of factors)
# p: estimated or calibrated probability (matrix, 1 column for each class)

# OUTPUT
# ------
# Result of the measure

#funI <- function(a,b){
#	if(a>b)
#		sol<-1.0
#	else
#		if(a<b)
#			sol<-0.0
#		else
#			sol<-0.5
#	return(sol)
#}

#funI <- function(a,b){
#	if(a>b){
#		sol<-1.0
#	}
#	if(a<b){
#		sol<-0.0
#	}
#	if(a==b){
#		sol<-0.5
#	}
#	
#	return(sol)
#}


#AUC2C <- function(yJ,yK,pJ,pK){
#	aux1<-0.0
#	for(i in 1:length(yJ)){
#		aux2<-0.0
#		for(t in 1:length(yJ)){
#			aux2<-aux2 + yK[t]*funI(pJ[i],pJ[t])
#		}
#		aux1<-aux1 + yJ[i]*aux2
#	}
#	return(aux1/(sum(yJ)*sum(yK)))
#}

#AUC2C <- function(yJ,yK,pJ,pK){
#	aux1<-0.0
#	for(i in 1:length(pJ)){
#		aux2<-0.0
#		for(t in 1:length(pJ)){
#			aux2<-aux2 + yK[t]*funI(pJ[i],pJ[t])
#		}
#		aux1<-aux1 + yJ[i]*aux2
#	}
#	return(aux1/(sum(yJ)*sum(yK)))
#}


#AU1U <- function(y,p){
#	classes <- levels(y)
#	noClass <- length(classes)
#	#m<-{}
#	#for(i in 1:noClass){
#	#	m<-c(m,length(which(y==classes[i])))
#	#}
#	len <- length(y)
#	auc<-0.0
#	for(j in 1:noClass){
#		for(k in 1:noClass){
#			if(k!=j){
#				yJ <- yK <- rep(0,len)
#				yJ[y==classes[j]] <- 1
#				yK[y==classes[k]] <- 1
#
#				pJ <- p[,j]
#				pK <- p[,k]
#		
#				auc<-auc+AUC2C(yJ,yK,pJ,pK)
#			}
#		}
#	}
#	auc<-auc/(noClass*(noClass-1))
#	return(auc)
#}

AUC2class <- function(y1,p1){
	p1o <- p1[order(p1,decreasing=TRUE)]
	y1o <- y1[order(p1,decreasing=TRUE)]
	len <- length(y1o)
	nPos <- sum(y1o)
	nNeg <- len - nPos
	
	auxP <- p1o
	auxY <- y1o
	yPos <- {}
	yNeg <- {}
	while(length(auxP)>0){
		repes <- which(auxP==auxP[1])
		sumaAux<-sum(auxY[repes])
		yPos<-c(yPos,sumaAux)
		yNeg<-c(yNeg,length(repes)-sumaAux)
		auxP <- auxP[-repes]
		auxY <- auxY[-repes]	
	}
	medPos <- 1/nPos
	medNeg <- 1/nNeg
	resAUCL <- 0
	quedanNeg <- nNeg
	for(i in 1:length(yNeg)){
		if(yNeg[i]!=0){
			resAUCL <- resAUCL + 0.5*yPos[i]*medPos*yNeg[i]*medNeg	
		}
		quedanNeg <- quedanNeg-yNeg[i]
		resAUCL <- resAUCL + quedanNeg*medNeg*yPos[i]*medPos	
	}
	return(resAUCL)		
}

AUC1vs1 <- function(y,p){
	classes <- levels(y)
	noClass <- length(classes)
	len <- length(y)
	
	resAUC <- {}
	for(i in 1:(noClass-1)){
		for(j in (i+1):noClass){
			yI <- rep(0,len)
			yI[y==classes[i]] <- 1
			yI <- yI[y==classes[i] | y==classes[j]]
			
			#yJ <- rep(0,len)
			#yJ[y==classes[j]] <- 1
			#yJ <- yJ[y==classes[i] | y==classes[j]]
			
			pT<-p[(y==classes[i] | y==classes[j]),c(i,j)]
			pT<-normaliseP(pT)
			
			resAUC <- c(resAUC,AUC2class(yI,pT[,1])) #+ AUC2class(yJ,pT[,2])
		}	
	}	
	#resAUC <- resAUC/(noClass*(noClass-1))
	
	resI<-0
	if(length(which(is.nan(resAUC)==FALSE))>0){
		resI<-mean(resAUC,na.rm=TRUE)
	}
	else{
		#print(paste("El dataset",indiceDS,"hay que descartarlo"))
		resI<-0.5
	}
	return(resI)
}

#y<-c(3,1,3,2,2)
#y<-as.factor(y)
#p<-matrix(c(0.1,0.3,0.6,0.5,0.2,0.3,0.4,0.1,0.5,0.2,0.6,0.2,0.3,0.4,0.3),ncol=3,byrow=TRUE)
AUC1vsAll <- function(y,p){
	classes <- levels(y)
	noClass <- length(classes)
	len <- length(y)
	
	resAUC <- {}
	for(i in 1:noClass){
		yI <- rep(0,len)
		yI[y==classes[i]] <- 1
			
		resAUC <- c(resAUC,AUC2class(yI,p[,i])) #+ AUC2class(yJ,pT[,2])
	}	
	#resAUC <- resAUC/(noClass*(noClass-1))
	
	resI<-0
	if(length(which(is.nan(resAUC)==FALSE))>0){
		resI<-mean(resAUC,na.rm=TRUE)
	}
	else{
		#print(paste("El dataset",indiceDS,"hay que descartarlo"))
		resI<-0.5
	}
	return(resI)
}


# Brier Score / Squared Loss:
#LossBS <- function(y,p){
#	yI<-as.integer(y)
#	yI[yI==2]<-0
#	return(sum((yI-p[,1])**2)/length(yI))
#	#return(sum(y*(1-p)**2+(1-y)*p**2)/length(y))
#}

# Log loss:
#LossLog <- function(y,p){
#	yI<-as.integer(y)
#	yI[yI==2]<-0
#	return(sum((-1)*yI*log(p[,1])-(1-yI)*log(1-p[,1]))/length(yI))
#}

# Exponential loss / Boosting loss:
#LossExp <- function(y,p){
#	yI<-as.integer(y)
#	yI[yI==2]<-0
#	return(sum(yI*sqrt((1-p[,1])/p[,1])+(1-yI)*sqrt(p[,1]/(1-p[,1])))/length(yI))
#}

# Calibration loss:
#LossC <- function(y,p){
#	yI<-as.integer(y)
#	yI[yI==2]<-0
#	auxP <- p[,1]
#	pj <- {}
#	nj <- {}
#	rj <- {}
#	while(length(auxP)>0){
#		pj <- c(pj,auxP[1])
#		repes <- which(auxP==auxP[1])
#		nj <- c(nj,length(repes))
#		rj <- c(rj,sum(yI[which(p[,1]==auxP[1])])/length(repes))
#		auxP <- auxP[-repes]	
#	}
#	return(sum(nj*(rj-pj)**2)/length(yI))	
#}

# Refinement loss:
#LossR <- function(y,p){
#	yI<-as.integer(y)
#	yI[yI==2]<-0
#	auxP <- p[,1]
#	pj <- {}
#	nj <- {}
#	rj <- {}
#	while(length(auxP)>0){
#		pj <- c(pj,auxP[1])
#		repes <- which(auxP==auxP[1])
#		nj <- c(nj,length(repes))
#		rj <- c(rj,sum(yI[which(p[,1]==auxP[1])])/length(repes))
#		auxP <- auxP[-repes]	
#	}
#	return(sum(nj*rj*(1-rj))/length(yI))	
#}


#y<-c("A","B","B","B","A","C","C","A","B","A","A","A","B","C","C","A","A","B","C","C")
#y<-as.factor(y)
#p<-matrix(c(0.5,0.3,0.2,0.3,0.6,0.1,0.4,0.5,0.1,0.1,0.8,0.1,0.6,0.1,0.3,0.2,0.4,0.4,0.3,0.3,0.4,0.7,0.2,0.1,0.1,0.5,0.4,0.5,0.1,0.4,0.4,0.3,0.3,0.7,0.1,0.2,0.2,0.5,0.3,0.2,0.1,0.7,0.1,0.1,0.8,0.7,0.2,0.1,0.5,0.2,0.3,0.3,0.5,0.2,0.4,0.1,0.5,0.2,0.3,0.5),ncol=3,byrow=TRUE)

# Calibration by Overlapping Bins (CalBin):
CalBin <- function(y,p){
	classes <- levels(y)
	noClass <- length(classes)
	len <- length(y)
	bin<-ceiling(len/10)
	cB<-{}
	for(i in 1:noClass){
		yI <- rep(0,len)
		yI[y==classes[i]] <- 1
		pI <- p[,i]
		# Sort by estimated probability decreasingly
		pInd <- order(pI,decreasing=TRUE)
		yOrd <- yI[pInd]
		pOrd <- pI[pInd]
		
		auxP <- pOrd
		pj <- {}
		nj <- {}
		rj <- {}
		while(length(auxP)>0){
			pj <- c(pj,auxP[1])
			repes <- which(auxP==auxP[1])
			nj <- c(nj,length(repes))
			rj <- c(rj,sum(yOrd[which(pOrd==auxP[1])])/length(repes))
			auxP <- auxP[-repes]	
		}
		newY<-{}
		newP<-{}
		for(j in 1:length(pj)){
			newY<-c(newY,rep(pj[j],nj[j]))
			newP<-c(newP,rep(rj[j],nj[j]))
		}
		
		yA<-pA<-{}
		for(j in 1:(len-bin+1)){
			yA<-c(yA,mean(newY[j:(j+bin-1)]))
			pA<-c(pA,mean(newP[j:(j+bin-1)]))
		}
		cB<-c(cB,mean(abs(yA-pA)))
	}
	return(mean(cB))	
}

# Brier Score / Squared Loss: (Multi-class)
LossBS <- function(y,p){
	classes <- levels(y)
	noClass <- length(classes)
	len <- length(y)
	bs<-{}
	for(i in 1:noClass){
		yI <- rep(0,len)
		yI[y==classes[i]] <- 1
		pI <- p[,i]
		bs<-c(bs,sum((yI-pI)**2)/length(yI))
	}
	return(mean(bs))	
}

# Calibration loss: (Multi-class)
LossC <- function(y,p){
	classes <- levels(y)
	noClass <- length(classes)
	len <- length(y)
	cl<-{}
	for(i in 1:noClass){
		yI <- rep(0,len)
		yI[y==classes[i]] <- 1
		#pI <- p[,i]
		
		auxP <- p[,i]
		pj <- {}
		nj <- {}
		rj <- {}
		while(length(auxP)>0){
			pj <- c(pj,auxP[1])
			repes <- which(auxP==auxP[1])
			nj <- c(nj,length(repes))
			rj <- c(rj,sum(yI[which(p[,i]==auxP[1])])/length(repes))
			auxP <- auxP[-repes]	
		}
		cl<-c(cl,sum(nj*(rj-pj)**2)/length(yI))
	}
	return(mean(cl))	
}

# Refinement loss: (Multi-class)
LossR <- function(y,p){
	classes <- levels(y)
	noClass <- length(classes)
	len <- length(y)
	cl<-{}
	for(i in 1:noClass){
		yI <- rep(0,len)
		yI[y==classes[i]] <- 1
		#pI <- p[,i]
		
		auxP <- p[,i]
		pj <- {}
		nj <- {}
		rj <- {}
		while(length(auxP)>0){
			pj <- c(pj,auxP[1])
			repes <- which(auxP==auxP[1])
			nj <- c(nj,length(repes))
			rj <- c(rj,sum(yI[which(p[,i]==auxP[1])])/length(repes))
			auxP <- auxP[-repes]	
		}
		cl<-c(cl,sum(nj*rj*(1-rj))/length(yI))
	}
	return(mean(cl))	
}

#cReal<-yTest
#cPred<-c0.NB
# cReal es factor y cPred es integer (1,2,...)
ACC <- function(cReal,cPred){
	return(length(which((as.integer(cReal)-cPred)==0))/length(cPred))
}

# cReal es factor y cPred es integer (1,2,...)
ACCc <- function(cReal,cPred,vClass,numClass){
	vAcc<-{}
	for(i in 1:numClass){
		indices<-which(cReal==vClass[i])
		vRealAux<-cReal[indices]
		vPredAux<-cPred[indices]
		vAcc<-c(vAcc,length(which((as.integer(vRealAux)-vPredAux)==0))/length(vPredAux))
	}
	
	return(vAcc)
}


## Implementacion de accuracy para 2 clases del paper de Flach & Lachiche
#ACC2c <- function(mScore,vReal){
#	cls<-levels(vReal)
#	auxP<-mScore[,1]
#	largo<-length(auxP)
#	pj<-{}
#	nj<-{}
#	posj<-{}
#	negj<-{}
#	cont<-0
#	while(length(auxP)>0){
#		cont <- cont+1
#		pj <- c(pj,auxP[1])
#		repes <- which(auxP==auxP[1])
#		nj <- c(nj,length(repes))
#		posj <- c(posj,length(which(vReal[repes]==cls[1])))
#		negj <- c(negj,nj[cont]-posj[cont])
#		auxP <- auxP[-repes]
#	}
#	indiceOrden<-order(pj,decreasing=TRUE)
#	pj<-pj[indiceOrden]
#	nj<-nj[indiceOrden]
#	posj<-posj[indiceOrden]
#	negj<-negj[indiceOrden]
#	
#	optimum<-sum(negj)/largo
#	current<-optimum
#	th<-pj[1]
#	for(i in 1:(length(pj)-1)){
#		current<-sum(posj[1:i])+sum(negj[(i+1):length(pj)])
#		if(i<length(pj)){
#			for(a2 in (i+1):length(pj)){
#				aux2<-negj[a2]
#			}
#		}	
#	}
#}

#scores<-newF
#class<-iClass[p]
#actual.classes<-y
#predicted.classes<-q
#vClass<-vClass
#numClass<-numClass

findBestWeight <- function(scores,class,actual.classes,predicted.classes,vClass,numClass){
	y<-actual.classes
	q<-predicted.classes
	auxP<-scores
	optimum<-ACC(y,q)
	current<-optimum
	pj<-{}
	nj<-{}
	rj<-{}
	ordenY<-{}
	iniY<-{}
	finY<-{}
	cont<-0
	pos<-0
	while(length(auxP)>0){
		pj <- c(pj,auxP[1])
		repes <- which(auxP==auxP[1])
		repesO <- which(scores==auxP[1])
		ordenY<-c(ordenY,repesO)
		nj <- c(nj,length(repes))
		cont<-cont+1
		iniY<-c(iniY,cont)
		pos<-pos+1
		cont<-iniY[pos]+nj[pos]-1
		finY<-c(finY,cont)
		for(i in 1:numClass){
			rj<-c(rj,length(which(q[repesO]==i)))
		}
		auxP <- auxP[-repes]
	}
	rj<-matrix(rj,ncol=numClass,byrow=TRUE)

	indiceOrden<-order(pj,decreasing=TRUE)
	pj<-pj[indiceOrden]
	nj<-nj[indiceOrden]
	rj<-rj[indiceOrden,]
	rj<-matrix(rj,ncol=numClass)
	iniY<-iniY[indiceOrden]
	finY<-finY[indiceOrden]
	
	yOrd<-{}
	for(i in 1:pos){
		yOrd<-c(yOrd,y[ordenY[iniY[i]:finY[i]]])
	}
	
	best.weight<-pj[1]
	cAux<-{}
	for(i in 1:length(pj)){
		for(j in 1:numClass){
			cAux<-c(cAux,rep(j,rj[i,j]))	
		}	
	}
	inicio<-0
	for(i in 1:length(nj)){
		inicio<-inicio+nj[i]
		#for(j in 1:inicio){
#			q[1:inicio]<-class
		cAux[1:inicio]<-class
		#}
		current<-ACC(yOrd,cAux)
		if(current>optimum){
			optimum<-current
			if((i+1)<=length(nj)){
				best.weight<-sqrt(pj[i]*pj[i+1])
			}
			else{
				best.weight<-pj[i]
			}	
		}	
	}
	
	return(best.weight)
}

#actual.classes<-yTrainVal
#scores<-p1TV.NB
#predicted.classes<-c1TV.NB
#vClass<-vClass
#numClass<-numClass
setWeights<-function(actual.classes,scores,predicted.classes,vClass,numClass){
	y<-actual.classes
	largoY<-length(y)
	f<-scores
	propC<-{}
	for(i in 1:numClass){
		propC<-c(propC,length(which(y==vClass[i])))
	}
	propC<-propC/largoY
	
	iClass<-1:numClass
	iClass<-iClass[order(propC,decreasing=TRUE)]
	#classesO<-vClass[iClass]
	
	w<-rep(0.0,numClass)
	w[iClass[1]]<-1.0
	for(p in 2:numClass){
		q<-{}
		newF<-{}
		fallos<-{}
		for(i in 1:largoY){
			iQ<-as.integer(which.max(w*f[i,]))
			miAux<-f[i,iClass[p]]/(w[iQ]*f[i,iQ])
			if(is.nan(miAux) | is.infinite(miAux)){
#				miAux<-0.0
				fallos<-c(fallos,i)
			}
			else{
				newF<-c(newF,miAux)
				q<-c(q,iQ)
			}
		}
		if(length(fallos)>0){
			w[iClass[p]]<-findBestWeight(newF,iClass[p],y[-fallos],q,vClass,numClass)
#			if(w[iClass[p]]==0){
#				w[iClass[p]]<-1
#			}
		}
		else{
			w[iClass[p]]<-findBestWeight(newF,iClass[p],y,q,vClass,numClass)
		}
#		if(w[iClass[p]]==0){
#			w[iClass[p]]<-1
#		}
	}
	if(sum(w)==1){ #Son todos ceros menos el primero
		w[which(w==0)]<-1
	}
	return(w)	
}
